/* 
 * File:         glpreview.c
 * 
 * Description:  handles rendering, opengl stuff, selection testing, etc
 * 
 * 
 * This source code is part of kludge3d, and is released under the 
 * GNU General Public License.
 * 
 * 
 */
 

/*
 * Portions of this file are from:
 * 
 * ME3D 3-D Modeler Program
 * Copyright (C) 1998 Sam Revitch
 *
 * G3d modeler program
 * Copyright (C) 2000 Blake E Hegerle
 *
 */


#include <math.h>

#include "tri.h"
#include "globals.h"
#include "frame.h"
#include "view.h"
#include "glpreview.h"

#include "memwatch.h"



// DEFINES, etc. ////////////////////////////////////////////////////////////


#define GIRDER_COLOR			0.8, 0.8, 0.0
#define WIREFRAME_COLOR			1.0, 1.0, 1.0
#define VTX_COLOR_NORMAL		0.5, 0.0, 0.0
#define VTX_COLOR_NORMAL_SEL		1.0, 0.0, 0.0




// PROTOTYPES ///////////////////////////////////////////////////////////





void glp_primitive_render_edit( model_view *view, Tri* t );

void glp_draw_cursor( model_view *view, GLdouble x, GLdouble y,
                      GLdouble z, GLdouble a );
void glp_draw_vertices( model_view *view, GList * points );
void glp_draw_girder( model_view *view, Point *p1, Point *p2 );

gboolean glp_draw_surfaces_tree( GNode *node, gpointer data );
void glp_draw_mesh( model_view *view, Mesh * m, gint box_it );
//void glp_draw_normals( model_group *group );
void glp_draw_grid( GLfloat *color, GLfloat *mins, GLfloat *maxes, GLfloat *spacing );




// GLOBALS ////////////////////////////////////////////////////////////////


//tex *active_texture = NULL;




///////////////////////////////////////////////////////////////////////////
//
// texture funcs
//
///////////////////////////////////////////////////////////////////////////


void texenv( void ) {
    GLfloat env_color[] = {0, 1, 0, 1};
    GLfloat border_color[] = {1, 0, 0, 1};
 
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
                     GL_NEAREST );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );
    glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL );
    glTexEnvfv( GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, env_color );
    glTexParameterfv( GL_TEXTURE_2D, GL_TEXTURE_BORDER_COLOR, border_color );
    glEnable( GL_TEXTURE_2D );
}


void tex_make_current( tex *ct ) {
    glTexImage2D( GL_TEXTURE_2D, 0, ct->depth, ct->width, ct->height, 0,
                  ct->depth == 4 ? GL_RGBA : GL_RGB, GL_UNSIGNED_BYTE,
                  ct->data );
}





//////////////////////////////////////////////////////////////////////////
//
//
// init/setup stuff
//
//
//////////////////////////////////////////////////////////////////////////

void glp_glconfigure( model_view *view ) {
    GLfloat whitelight[] = { 1.0, 1.0, 1.0, 0.5 };
    GLfloat amblight[] = { 0.2, 0.2, 0.1, 1.0 };
    GLfloat lightpos[] = { -5.0, 5.0, -2.0, 1.0 };

    g_return_if_fail( view != NULL );

// a test...
//g_return_if_fail( gtk_gl_area_make_current( GTK_GL_AREA( view->glarea ) ) );

    /* clear frame color */
    glClearColor( view_config.bkgnd_color[ 0 ], view_config.bkgnd_color[ 1 ],
                  view_config.bkgnd_color[ 2 ], 0.0 );

    /* These are the same under all circumstances */
    glEnable( GL_DEPTH_TEST );
    glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );

    /* lighting is the same (for now) */
    glEnable( GL_LIGHT0 );
    glLightfv( GL_LIGHT0, GL_POSITION, lightpos );
    glLightfv( GL_LIGHT0, GL_DIFFUSE, whitelight );
    glLightfv( GL_LIGHT0, GL_AMBIENT, amblight );

    switch( view_config.triangle_order ) {
    case MP_TRI_ORDER_CW:
        glEnable( GL_CULL_FACE );
        glFrontFace( GL_CW );
        break;
    case MP_TRI_ORDER_CCW:
        glEnable( GL_CULL_FACE );
        glFrontFace( GL_CCW );
        break;
    case MP_TRI_ORDER_ALL:
        glDisable( GL_CULL_FACE );
        break;
    default:
        break;
    }

//FIXME
//    if ( view->pDoc->config->smooth_mode )
//        glShadeModel( GL_SMOOTH );
//    else
        glShadeModel( GL_FLAT );

    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();


}


void glp_glviewport( model_view *view ) {
    /* Set the camera */
    if ( GTK_WIDGET_REALIZED( GTK_WIDGET( view->glarea ) ) &&
            ( GTK_WIDGET( view->glarea ) ->allocation.width > 1 ) &&
            ( GTK_WIDGET( view->glarea ) ->allocation.height > 1 ) ) {
        GLfloat xd, yd, hwr, fov_rad;

        g_return_if_fail( gtk_gl_area_make_current( GTK_GL_AREA( view->glarea ) ) );

        glViewport( 0, 0, GTK_WIDGET( view->glarea ) ->allocation.width,
                    GTK_WIDGET( view->glarea ) ->allocation.height );
        hwr = ( GLfloat ) GTK_WIDGET( view->glarea ) ->allocation.height /
              GTK_WIDGET( view->glarea ) ->allocation.width;

        fov_rad = ( view_config.fov * M_PI ) / 360;
        xd = view_config.near * tan( fov_rad );
        yd = xd * hwr;
        glMatrixMode( GL_PROJECTION );
        glLoadIdentity();
        glFrustum( -xd, xd, -yd, yd, view_config.near, view_config.far );
        glMatrixMode( GL_MODELVIEW );

#ifdef VERBOSE_DEBUG
        g_print( "glp_glviewport %ux%u %f->%f (%f) %f %f %f\n",
                 GTK_WIDGET( view->glarea ) ->allocation.width,
                 GTK_WIDGET( view->glarea ) ->allocation.height,
                 view_config.near, view_config.far, view_config.fov,
                 xd, yd, hwr );
#endif
    }
}

void glp_ortho_viewport( model_view *view ) {
    
    // sets the extents for the ortho graphic views...
    
    float halfdist;
    float aspect_ratio = 1.0;
    
    
    halfdist = (view->dist / 2.0);
    
    view->minX = - halfdist;
    view->maxX = halfdist;

    view->minY = - halfdist;
    view->maxY = halfdist;

    view->minZ = - halfdist;
    view->maxZ = halfdist;
    
    if( view->glarea != NULL ) {

        aspect_ratio = (float)(GTK_WIDGET( view->glarea )->allocation.width) / 
                       (float)(GTK_WIDGET( view->glarea )->allocation.height);

        switch( view->type ) {
            case FRAME_XY:
                view->minX = view->minY * aspect_ratio;
                view->maxX = view->maxY * aspect_ratio;
                break;
            case FRAME_XZ:
                view->minX = view->minZ * aspect_ratio;
                view->maxX = view->maxZ * aspect_ratio;
                break;
            case FRAME_YZ:
                view->minY = view->minZ * aspect_ratio;
                view->maxY = view->maxZ * aspect_ratio;
                break;
            default:
                break;
        }

    }
    
}


void glp_setup_view( model_view *view ) {
    // because we're using multiple views, each with it's own 
    // camera location and projection type, we need to set up the 
    // view before every render cycle.  This involves lots of state 
    // changes, which is (as I understand it) a costly operation.
    // Is there a better way to do this?
    
    glPushMatrix();
    glTranslatef( 0, 0, - ( view->dist + view_config.near ) );
    glPushMatrix();
    glRotatef( -( view->pitch ), 1.0, 0.0, 0.0 );
    glPushMatrix();
    glRotatef( view->heading, 0.0, 0.0, 1.0 );
    glPushMatrix();
    glTranslatef( view->camX, view->camY, view->camZ  );
    

//fixme - need to set viewport size too?  ortho views are different size from 3d view...
    if( view->type != FRAME_3D ) {
        // switch to projection mode
        glMatrixMode(GL_PROJECTION);

        // save previous matrix which contains the 
        //settings for the perspective projection
        glPushMatrix();

        // reset matrix
        glLoadIdentity();

        // set a 2D orthographic projection
        switch( view->type ) {
            case FRAME_XY:
                glOrtho( view->minX, view->maxX, view->minY, view->maxY, view_config.far, -view_config.far );
                break;
            case FRAME_XZ:
                glOrtho( view->minX, view->maxX, view->minZ, view->maxZ, view_config.far, -view_config.far );
                break;
            case FRAME_YZ:
                glOrtho( view->minY, view->maxY, view->minZ, view->maxZ, view_config.far, -view_config.far );
                break;
            default:
                break;
        }
        


        glMatrixMode(GL_MODELVIEW);
    
    }
}


void glp_cleanup_view( model_view *view ) {
    // should be called after every call to glp_setup_view()

    if( view->type != FRAME_3D ) {
        glMatrixMode(GL_PROJECTION);
        glPopMatrix();
        glMatrixMode(GL_MODELVIEW);
    }
    
    glPopMatrix();
    glPopMatrix();
    glPopMatrix();
    glPopMatrix();
    
}






//////////////////////////////////////////////////////////////////////////
//
//
//  rendering stuff
//
//
//////////////////////////////////////////////////////////////////////////



// used only in test_click_surface... eventually will be phased out
void glp_primitive_render( model_view *view, Tri* t ) {
    
    g_return_if_fail( GTK_WIDGET_VISIBLE( view->gltable ) );

    g_return_if_fail( t != NULL );

    glBegin( GL_TRIANGLES );
    
    glVertex3dv( t->vertices[ 0 ]->v );
    glVertex3dv( t->vertices[ 1 ]->v );
    glVertex3dv( t->vertices[ 2 ]->v );

    glEnd();
}


void glp_draw_primitives( model_view *view, GList* tris ) {
    
    Tri* t = NULL;
    
    g_return_if_fail( GTK_WIDGET_VISIBLE( view->gltable ) );
    g_return_if_fail( tris != NULL );

    glBegin( GL_TRIANGLES );
    
    for( ; tris != NULL; tris = tris->next ) {
        t = (Tri*)tris->data;
        glColor3f( WIREFRAME_COLOR );
        glVertex3dv( t->vertices[ 0 ]->v );
        glColor3f( WIREFRAME_COLOR );
        glVertex3dv( t->vertices[ 1 ]->v );
        glColor3f( WIREFRAME_COLOR );
        glVertex3dv( t->vertices[ 2 ]->v );
    }

    glEnd();
}


void glp_draw_primitives_textured( model_view *view, GList * tris, 
                                   float texRepX, float texRepY ) {
    
    Tri* t = NULL;
    
    g_return_if_fail( GTK_WIDGET_VISIBLE( view->gltable ) );
    g_return_if_fail( tris != NULL );
//    g_return_if_fail( active_texture != NULL );
    
    if( texRepX == 0.0 ) texRepX = 1.0;
    if( texRepY == 0.0 ) texRepY = 1.0;
    
    glEnable( GL_TEXTURE_2D );
    glBegin( GL_TRIANGLES );

    for( ; tris != NULL; tris = tris->next ) {
        t = (Tri*)tris->data;
        glTexCoord2d( t->tv[ 0 ].x * texRepX,
                      t->tv[ 0 ].y * texRepY );
        glVertex3dv( t->vertices[ 0 ]->v );
        glTexCoord2d( t->tv[ 1 ].x * texRepX,
                      t->tv[ 1 ].y * texRepY );
        glVertex3dv( t->vertices[ 1 ]->v );
        glTexCoord2d( t->tv[ 2 ].x * texRepX,
                      t->tv[ 2 ].y * texRepY );
        glVertex3dv( t->vertices[ 2 ]->v );
    }

    glEnd();
    glDisable( GL_TEXTURE_2D );

}


void glp_primitive_render_edit( model_view *view, Tri* t ) {

    g_return_if_fail( GTK_WIDGET_VISIBLE( view->gltable ) );
    g_return_if_fail( t != NULL );

    if( t->vertices[ 0 ] != NULL ) {
        if( t->vertices[ 1 ] != NULL ) {
            glp_draw_girder( view, t->vertices[ 0 ], t->vertices[ 1 ] );
            if( t->vertices[ 2 ] != NULL ) {
                glp_draw_girder( view, t->vertices[ 1 ], t->vertices[ 2 ] );
                glp_draw_girder( view, t->vertices[ 2 ], t->vertices[ 0 ] );
            } else {
                glp_draw_cursor( view, t->vertices[ 1 ]->v[0], t->vertices[ 1 ]->v[1], t->vertices[ 1 ]->v[2], 0 );
            }
        } else {
            glp_draw_cursor( view, t->vertices[ 0 ]->v[0], t->vertices[ 0 ]->v[1], t->vertices[ 0 ]->v[2], 0 );
        }
    } 
    

}



void glp_draw_cursor( model_view *view, GLdouble x, GLdouble y,
                      GLdouble z, GLdouble a ) {
    glPushMatrix();
    glTranslatef( x, y, z );
    glPushMatrix();
    glRotatef( a, 0, 1, 0 );

    glInterleavedArrays( GL_C4F_N3F_V3F, 0, cursor_glarray );
    glDrawArrays( GL_TRIANGLES, 0, 24 );

    glPopMatrix();
    glPopMatrix();

}




void glp_draw_vertices( model_view *view, GList* points ) {

    glInterleavedArrays( GL_N3F_V3F, 0, vertex_glarray );

    while ( points != NULL ) {
        Point * p = (Point*)points->data;

        if ( p->selected )
            glColor3f( VTX_COLOR_NORMAL_SEL );
        else 
            glColor3f( VTX_COLOR_NORMAL );

        glPushMatrix();
        glTranslatef( p->v[ 0 ], p->v[ 1 ], p->v[ 2 ] );
        glDrawArrays( GL_QUADS, 0, 24 );
        glPopMatrix();

        points = points->next;
    }

}




void glp_draw_girder( model_view * view, Point * p1, Point * p2 ) {

    GLfloat length, v1[ 3 ], n[ 2 ], sin_theta, theta;

    if ( p1->v[ 2 ] > p2->v[ 2 ] ) {	    	/* +/- issue with arcsin */
        Point *temp = p2;
        p2 = p1;
        p1 = temp;
    }					/* what a pain */

    v1[ 0 ] = p2->v[ 0 ] - p1->v[ 0 ];
    v1[ 1 ] = p2->v[ 1 ] - p1->v[ 1 ];
    v1[ 2 ] = p2->v[ 2 ] - p1->v[ 2 ];

    sin_theta = ( v1[ 0 ] * v1[ 0 ] ) + ( v1[ 1 ] * v1[ 1 ] );
    length = sqrt( sin_theta + ( v1[ 2 ] * v1[ 2 ] ) );
    sin_theta = sqrt( sin_theta );
    if ( sin_theta == 0 ) {
        n[ 0 ] = 1; 		/* doesn't matter */
        n[ 1 ] = 0;
    } else {
        n[ 0 ] = -v1[ 1 ] / sin_theta;
        n[ 1 ] = v1[ 0 ] / sin_theta;
    }
    sin_theta = sin_theta / length;
    theta = asin( sin_theta ) * ( ( GLdouble ) 180 / M_PI );
#ifdef VERBOSE
    printf( "v = <%f, %f, %f> ", v1[ 0 ], v1[ 1 ], v1[ 2 ] );
    printf( "n = <%f, %f, 0> theta = %f\n", n[ 0 ], n[ 1 ], theta );
#endif

    /* Set the remap matrix appropriately */
    glPushMatrix();
    glTranslated( p1->v[ 0 ], p1->v[ 1 ], p1->v[ 2 ] );
    glPushMatrix();
    glRotatef( theta, n[ 0 ], n[ 1 ], 0 );

    girder_glarray[ 17 ] = length;                /* Patch the array */
    girder_glarray[ 23 ] = length;
    girder_glarray[ 41 ] = length;
    girder_glarray[ 47 ] = length;
    girder_glarray[ 65 ] = length;
    girder_glarray[ 71 ] = length;
    girder_glarray[ 89 ] = length;
    girder_glarray[ 95 ] = length;

    glColor3f( GIRDER_COLOR );
    glInterleavedArrays( GL_N3F_V3F, 0, girder_glarray );
    glDrawArrays( GL_QUADS, 0, 16 );

    glPopMatrix();
    glPopMatrix();

}




gboolean glp_draw_surfaces_tree( GNode *node, gpointer data ) {
    // recurse through tree
    
    Mesh* m = (Mesh*)node->data;

    glp_draw_mesh( (model_view*)data, m, 0 );
    
    return FALSE;  /* we want it to continue... */
}


void glp_draw_mesh( model_view *view, Mesh * m, gint box_it ) {

    static GLfloat initializer[ 4 ] = { 0.0, 0.0, 0.0, 1.0 };
    static float default_diff[4] = { 0.8, 0.8, 0.8, 1.0 };
    static float default_amb[4] = { 0.2, 0.2, 0.2, 1.0 };
    static float default_spec[4] = { 0.0, 0.0, 0.0, 1.0 };
    static float default_emis[4] = { 0.0, 0.0, 0.0, 1.0 };
    static float default_shininess = 0.0;
//    GList * triangles;
    
    if( m == NULL )
        return;
    
//    triangles = m->triangles;
    
    // hack...
//    if( active_texture != m->texture ) {
//        active_texture = m->texture;
//        if( active_texture != NULL )
//            tex_make_current( m->texture );
//    }



    glPushMatrix();
    if( m->location != NULL )
        glTranslatef( m->location->v[0], m->location->v[1], m->location->v[2] );
    

    if( m->triangles != NULL )
    switch ( view_config.polygon_mode ) {

    case MP_SHOW_TEXTURED:
        glEnable( GL_LIGHTING );
        glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
        
        if( m->texture != NULL ) {
            tex_make_current( m->texture );
            texenv();
            glp_draw_primitives_textured( view, m->triangles, m->texrep_x, m->texrep_y );
            break;
        }
        // ...else continue into the wireframe code below...
        
    case MP_SHOW_WIREFRAME:			/* wire frame mode is non-shaded */
        glColor3f( 1.0, 1.0, 1.0 );		/* unfortunately not much faster */
        glDisable( GL_LIGHTING );
        glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
        
        glp_draw_primitives( view, m->triangles );
        
        break;

    case MP_SHOW_FLAT:
        glEnable( GL_LIGHTING );
        glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
        glEnable( GL_COLOR_MATERIAL );
        glMaterialfv( GL_FRONT, GL_DIFFUSE, default_diff );
        glMaterialfv( GL_FRONT, GL_AMBIENT, default_amb );
        glMaterialfv( GL_FRONT, GL_SPECULAR, default_spec );
        glMaterialfv( GL_FRONT, GL_EMISSION, default_emis );
        glMaterialf( GL_FRONT, GL_SHININESS, default_shininess );
        glColorMaterial( GL_FRONT, GL_AMBIENT_AND_DIFFUSE );
        
        glp_draw_primitives( view, m->triangles );
        
        break;


    default:
        break;
    }


    glDisable( GL_LIGHTING );
    glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
    
    /* necessary for drawing vertices, etc */
    glEnable( GL_COLOR_MATERIAL );
    glColorMaterial( GL_FRONT, GL_AMBIENT_AND_DIFFUSE );
    glMaterialfv( GL_FRONT, GL_SPECULAR, initializer );
    glMaterialfv( GL_FRONT, GL_EMISSION, initializer );
    glMaterialf( GL_FRONT, GL_SHININESS, 0.0 );


    if( current_mesh == m ) {
        if ( view_config.vertices_visible ) 
            glp_draw_vertices( view, m->points );
        if ( selected_tri != NULL )
            glp_primitive_render_edit( view, selected_tri );
        if ( partially_completed_tri != NULL )  /* global from view.h */
            glp_primitive_render_edit( view, partially_completed_tri );
    }

    glPopMatrix();




}



void glp_draw_grid( GLfloat *color, GLfloat *mins, GLfloat *maxes, GLfloat *spacing ) {

    GLfloat coord;
    GLfloat multiple;

    glColor3fv( color );
    glBegin( GL_LINES );

    multiple = mins[0] / spacing[0];
    coord = spacing[0] * rint( multiple );
    for ( ; coord <= maxes[ 0 ]; coord += spacing[ 0 ] ) {
        glVertex3f( coord, 0, mins[ 2 ] );
        glVertex3f( coord, 0, maxes[ 2 ] );
        glVertex3f( coord, mins[ 1 ], 0 );
        glVertex3f( coord, maxes[ 1 ], 0 );
    }

    multiple = mins[1] / spacing[1];
    coord = spacing[1] * rint( multiple );
    for ( ; coord <= maxes[ 1 ]; coord += spacing[ 1 ] ) {
        glVertex3f( 0, coord, mins[ 2 ] );
        glVertex3f( 0, coord, maxes[ 2 ] );
        glVertex3f( mins[ 0 ], coord, 0 );
        glVertex3f( maxes[ 0 ], coord, 0 );
    }

    multiple = mins[2] / spacing[2];
    coord = spacing[2] * rint( multiple );
    for ( ; coord <= maxes[ 2 ]; coord += spacing[ 2 ] ) {
        glVertex3f( 0, mins[ 1 ], coord );
        glVertex3f( 0, maxes[ 1 ], coord );
        glVertex3f( mins[ 0 ], 0, coord );
        glVertex3f( maxes[ 0 ], 0, coord );
    }

    glEnd();

}




void glp_render( model_view *view ) {

    if ( view == NULL )
        return;
    
    if ( !gtk_gl_area_make_current( GTK_GL_AREA( view->glarea ) ) )
        return; 

    glDisable( GL_LIGHTING );
    glColor3f( 1.0, 1.0, 1.0 );

    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );


    glp_setup_view( view );

    
    if ( view_config.grid_visible )
        glp_draw_grid( view_config.grid_color, view->grid_mins,
                       view->grid_maxes, view_config.grid_spacing );

    
    //render the surfaces for all meshes under current_node
    if( current_node != NULL )
        g_node_traverse( current_node, G_IN_ORDER, G_TRAVERSE_ALL, -1,
                         glp_draw_surfaces_tree, view );



    glp_cleanup_view( view );


    gtk_gl_area_swapbuffers( GTK_GL_AREA( view->glarea ) );
    
}


/////////////////////////////////////////////////////////////////////////
//
//
//  funcs for selection of primitives
//
//
//////////////////////////////////////////////////////////////////////////


Point *glp_test_click_vertex( model_view *view, 
			      GLfloat *z, guint x, guint y ) {

    GLdouble best_z = 9999999.0, click_r, cr1, cr2, cr3;
    GLfloat array[ 8 ];
    Point *best_point = NULL;
    GList *points = NULL;

    g_return_val_if_fail ( view != NULL, NULL );
    g_return_val_if_fail ( view->glarea != NULL, NULL );
    g_return_val_if_fail ( current_mesh != NULL, NULL );

    /* fragments for computing the approximate vertex radius in a frustum perspective */
    /* Ze = (near*far) / (far - Zwin*(far-near)) */
    /* but these fragments are to find the size of the pixel square for each vertex */
    /* i.e. winradius / (Ze * tan(fov/2)) */
    cr1 = ( view_config.vertex_rad * M_SQRT2 * view->glarea->allocation.width ) /
          ( 2 * view_config.near * view_config.far * tan( view_config.fov / 2 ) );
    cr2 = ( view_config.far );
    cr3 = ( view_config.far - view_config.near );
    /* the maximum vertex radius is sqrt(3)*VTX_RAD, the minimum is VTX_RAD */

    g_return_val_if_fail( gtk_gl_area_make_current( GTK_GL_AREA( view->glarea ) ), NULL );
    glFeedbackBuffer( 8, GL_3D, array );
    glRenderMode( GL_FEEDBACK );

    // need to flip y axis; X11 origin is upperleft, opengl origin is lowerleft
    y = view->glarea->allocation.height - y;

    glp_setup_view( view );
    
    glPushMatrix();
    if( current_mesh->location != NULL )
        glTranslatef( current_mesh->location->v[0], current_mesh->location->v[1], current_mesh->location->v[2] );
        
    points = current_mesh->points;
    

    while ( points != NULL ) {
        Point* p = (Point *)points->data;
        glBegin( GL_POINTS );
        glVertex3dv( p->v );
        glEnd();

        /*    printf("glRenderMode returns %i\n", glRenderMode(GL_FEEDBACK)); */

        if ( 4 != glRenderMode( GL_FEEDBACK ) ) {
            points = points->next;
            continue; 			/* Point isn't within clipping volume */
        }

        /* Compute click radius */
        click_r = cr1 * ( cr2 - ( cr3 * array[ 3 ] ) ); 	/* assemble fragments */

        array[ 1 ] = rint( array[ 1 ] ); 		/* round to nearest integer */
        array[ 2 ] = rint( array[ 2 ] );		/* normally float->int just */
        click_r = rint( click_r );			/* chops off the decimal point and beyond */

        /*    d = (view->near * view->far) /
        (doc->view - (array[3] * (view->far - view->near))); */

        /*    printf("vertex <%f,%f,%f> @ (%8.8f,%8.8f) r=%8.8f\n",
        points->v[0], points->v[1], points->v[2], array[1], array[2], click_r); */

        if ( ( array[ 1 ] + click_r >= x ) &&
             ( array[ 1 ] - click_r <= x ) &&
             ( array[ 2 ] + click_r >= y ) &&
             ( array[ 2 ] - click_r <= y ) &&
             ( array[ 3 ] < best_z ) ) {
            best_point = p;
            best_z = array[ 3 ];
        }

        points = points->next;
    }

    glp_cleanup_view( view );
    
    glPopMatrix();
    
    glRenderMode( GL_RENDER );
    *z = best_z;
    return best_point;
}



#define glp_ARRAY_SIZE 2048
Tri *glp_test_click_surface( model_view *view,
                             GLfloat *z, guint xc, guint yc ) {

    GLdouble best_z = 9999999.0, bz_vert, m, b, x, y;
    GLfloat array[ glp_ARRAY_SIZE ], *v1, *v2;
    Tri *best_tri = NULL;
    GList *triangles = NULL;
    int i, j, len, polylen, intersections;

    x = xc;
    y = yc;

    g_return_val_if_fail ( view != NULL, NULL );
    g_return_val_if_fail ( view->glarea != NULL, NULL );
    g_return_val_if_fail ( current_mesh != NULL, NULL );

    g_return_val_if_fail( gtk_gl_area_make_current( GTK_GL_AREA( view->glarea ) ), NULL );
    glFeedbackBuffer( glp_ARRAY_SIZE, GL_3D, array );
    glRenderMode( GL_FEEDBACK );

    // need to flip y axis; X11 origin is upperleft, opengl origin is lowerleft
    y = view->glarea->allocation.height - y;

    glp_setup_view( view );
    
    glPushMatrix();
    if( current_mesh->location != NULL )
        glTranslatef( current_mesh->location->v[0], current_mesh->location->v[1], current_mesh->location->v[2] );

    triangles = current_mesh->triangles;

    while ( triangles != NULL ) {
        Tri* t = (Tri*)triangles->data;
        glp_primitive_render( view, t );

        /* now process the feedback buffer and check for hits */
        len = glRenderMode( GL_FEEDBACK );
        intersections = 0;
        for ( i = 0; i < len; ) {
            g_return_val_if_fail( array[ i ] == GL_POLYGON_TOKEN, NULL );

            i++;
            polylen = ( int ) array[ i ];
            i++;

            bz_vert = 9999999.0;
            v1 = &( array[ 3 * ( polylen - 1 ) + i ] );
            for ( j = 0; j < polylen; j++ ) {
                v2 = v1;
                v1 = &( array[ i ] );
                i += 3;

                if ( array[ 2 ] < bz_vert )
                    bz_vert = array[ 2 ];

                /* try to eliminate the line represented by v1 and v2 */

                /* in order to intersect the test line, the v1<->v2 line */
                /* must have a vertex to both sides of the click point */
                if ( !( ( v1[ 0 ] < x ) && ( v2[ 0 ] > x ) ) &&
                        !( ( v2[ 0 ] < x ) && ( v1[ 0 ] > x ) ) )
                    continue;
                /* note that it's now impossible for v1[0] == v2[0] */

                /* the line can be immediately qualified if both are above */
                /* the click point */
                if ( ( v1[ 1 ] < y ) && ( v2[ 1 ] < y ) ) {
                    intersections++;
                    continue;
                }

                /* if both vertices are below the click point, the line */
                /* can't intersect */
                if ( ( v1[ 1 ] > y ) && ( v2[ 1 ] > y ) )
                    continue;

                /* now the situation is sticky, the line formula must be used */
                /* to test for intersection */

                m = ( v2[ 1 ] - v1[ 1 ] ) / ( v2[ 0 ] - v1[ 0 ] );
                b = v1[ 1 ] - ( m * v1[ 0 ] );

                if ( y < ( ( m * x ) + b ) )
                    continue;

                intersections++;
            }

            intersections &= 0x1;	/* an odd number of intersections means */
					/* it was clicked on */
            if ( intersections && ( bz_vert < best_z ) ) {
                best_tri = t;
                best_z = bz_vert;
            }
        }

        triangles = triangles->next;
    }	/* triangles != NULL */

    
    glp_cleanup_view( view );
    
    glPopMatrix();
    glRenderMode( GL_RENDER );
    *z = best_z;
    return best_tri;
}





///////////////////////////////////////////////////////////////////////////////
//
//
// ONLY DEAD CODE BELOW THIS POINT
//
//
///////////////////////////////////////////////////////////////////////////////



#if 0 
void glp_draw_normals( struct model_group *group ) {
    struct surface_link *pL;
    struct surface_l *pS;

    pS = group->surface_list;

    if ( pS == NULL )
        return ;

    glColor3f( 1.0, 1.0, 1.0 );
    glBegin( GL_LINES );

    while ( pS != NULL ) {
        pL = pS->sl_list;
        g_return_if_fail( pL != NULL );
        while ( pL != NULL ) {
            if ( ( pL->normal[ 0 ] != 0 ) || ( pL->normal[ 1 ] != 0 ) || ( pL->normal[ 2 ] != 0 ) ) {
                glVertex3dv( &( pL->vert->v[ 0 ] ) );
                glVertex3d( pL->vert->v[ 0 ] + pL->normal[ 0 ], pL->vert->v[ 1 ] + pL->normal[ 1 ],
                            pL->vert->v[ 2 ] + pL->normal[ 2 ] );
            }
            pL = pL->s_next;
        }
        pS = pS->pNext;
    }

    glEnd();
}





#endif



