/* 
 * File:         view.c
 * 
 * Description:  handles 3d views, vertex&triangle selection, etc
 * 
 * 
 * This source code is part of kludge3d, and is released under the 
 * GNU General Public License.
 * 
 * 
 */
 

#include <math.h>

#include "tri.h"
#include "toolbar.h"
#include "globals.h"
#include "frame.h"
#include "editor_models.h"
#include "tex_app.h"
#include "glpreview.h"
#include "view.h"


#include "memwatch.h"


// DEFINES /////////////////////////////////////////////////////////////

#define RCV_VIEWPORT    0x0001
#define RCV_PROPERTIES  0x0002

struct _mouse_state {
    int button_pressed;   /* is this button down? */
    int has_moved;        /* has the mouse been moved while this button was down? */
    int startx, starty;   /* starting location for a drag */
    int endx, endy;
    int view_index;       /* one of the view types, ie FRAME_XY, etc */
};


// PROTOTYPES ///////////////////////////////////////////////////////////


gboolean view_mouse_down( GtkWidget *widget, GdkEventButton *event,
                           model_view *view );
gboolean view_mouse_motion( GtkWidget *widget, GdkEventMotion *event,
                           model_view *view );
gboolean view_mouse_up( GtkWidget *widget, GdkEventButton *event,
                           model_view *view );

void view_pan( model_view* view, float deltaX, float deltaY, float deltaZ );
void view_zoom( model_view* view, float deltaX, float deltaY, float deltaZ );
void view_lasso_select( model_view *view, struct _mouse_state* ms );


void view_reconfigure( model_view *view, gint flags );

Point *glp_test_click_ortho( model_view *view, guint x, guint y );



// GLOBALS ////////////////////////////////////////////////////////////////

model_view *views[4];
model_view_config view_config;

int click_mode[ 4 ];

Tri* partially_completed_tri = NULL;

int changed_views = FALSE;

struct _mouse_state mouse_state[3];  /* one for each of the mouse buttons */

GLfloat *cursor_glarray = NULL;            /* GL arrays describing */
GLfloat *vertex_glarray = NULL;            /* helper objects */
GLfloat *girder_glarray = NULL;




//////////////////////////////////////////////////////////////////////////
//
//
//  callbacks, of all kinds
//
//
///////////////////////////////////////////////////////////////////////////



void view_glarea_realize( GtkWidget *widget, gpointer user_data ) {
    model_view * view = ( model_view * ) user_data;
    view_reconfigure( view, RCV_PROPERTIES | RCV_VIEWPORT );
}

gboolean view_glarea_configure( GtkWidget *widget, GdkEventConfigure *event,
                                gpointer user_data ) {
    model_view * view = ( model_view * ) user_data;
    view_reconfigure( view, RCV_VIEWPORT );
    glp_render( view );
    return TRUE;
}

gboolean view_glarea_expose( GtkWidget *widget, GdkEventExpose *event,
                             gpointer user_data ) {
    model_view * view = ( model_view * ) user_data;
    glp_render( view );
    return TRUE;
}




gint vadj_changed( GtkWidget *widget, gpointer data ) {
    views[FRAME_3D]->pitch = ( double ) GTK_ADJUSTMENT( widget ) ->value;
    gtk_widget_draw( views[FRAME_3D]->glarea, NULL );

//printf( "3d view pitch: %f\n", views[FRAME_3D]->pitch );
    return TRUE;
}


gint hadj_changed( GtkWidget *widget, gpointer data ) {
    views[FRAME_3D]->heading = ( double ) GTK_ADJUSTMENT( widget ) ->value;
    gtk_widget_draw( views[FRAME_3D]->glarea, NULL );

//printf( "3d view heading: %f\n", views[FRAME_3D]->heading );
    return TRUE;
}





gboolean view_mouse_leave( GtkWidget *widget, GdkEventCrossing *event,
                           model_view *view ) {

// if the cursor left the glarea, with a mouse button held down, then set changed_views
//    if( mouse_state[0].button_pressed || 
//        mouse_state[1].button_pressed || 
//        mouse_state[2].button_pressed )
        changed_views = TRUE;

//printf( "exited view %i, changed_views is %i\n", view->type, changed_views );
    return TRUE;
}





gboolean view_mouse_down( GtkWidget *widget, GdkEventButton *event,
                           model_view *view ) {

    int rerender = FALSE;
    Point* p = NULL;
    GLfloat z;

//FIXME - what if user changes current_mesh between clicks while adding a triangle?


//    if ( event->button == 3 ) {
//        glp_configure_menu( view );
//        docmenu_show( view, event );
//        return TRUE;
//    }
    
if( current_mesh != NULL )
    switch ( click_mode[ event->button ] ) {

    case CM_POINT_ADD:
        p = glp_test_click_ortho( view, event->x, event->y );

        if( p != NULL ) {
            if( current_mesh->location != NULL ) {
                p->v[0] -= current_mesh->location->v[0];
                p->v[1] -= current_mesh->location->v[1];
                p->v[2] -= current_mesh->location->v[2];
            }
            p->v[0] -= view->camX;  // see note in glp_test_click_ortho
            p->v[1] -= view->camY;
            p->v[2] -= view->camZ;
            if( current_node->parent != NULL ) {
                // IMPORTANT - we don't want to add points to the root mesh
//FIXME - this still needs work (users can paste points into root mesh, for example)
                point_add_v( current_mesh, p );
            } else {
                free( p );
                p = NULL;
            }
        }
        rerender = TRUE;
        
        break;

    case CM_POINT_DEL:
        
        p = glp_test_click_vertex( view, &z, event->x, event->y );
        point_del_v( current_mesh, p );
        /* delete any lines or triangles that use this point */
        //line_del_point( p );
        tris_del_point( p );
        /* we may have just deleted one of the selected points... */
        point_unselect_all();
        selected_tri = NULL;
        p = NULL;
        rerender = TRUE;
        break;

    case CM_TRI_ADD:
        
        if( partially_completed_tri == NULL ) {
            // create a new triangle
            partially_completed_tri = tri_new();
        }
        
        p = glp_test_click_vertex( view, &z, event->x, event->y );
        if( p != NULL ) {
            int i;
            for( i = 0; i < 3; i++ ) {
                // find the first unassigned point in partially_completed_tri
                if( partially_completed_tri->vertices[i] == NULL ){
                    partially_completed_tri->vertices[i] = p;
                    break;
                }
            }
            if( i >= 2 ) {
                // partially_completed_tri has points assigned to all three elements
                // time to add it to the model
                if( current_mesh != NULL )
                    tris_add_v( current_mesh, partially_completed_tri );
                partially_completed_tri = NULL;
//leak?
            }
            rerender = TRUE;
        }
        
        break;

    case CM_TRI_SEL:
        selected_tri = glp_test_click_surface( view, &z, event->x, event->y );
        rerender = TRUE;
        break;

    case CM_TRI_TEX:
        selected_tri = glp_test_click_surface( view, &z, event->x, event->y );
        show_tex_app();
        rerender = TRUE;
        break;

    case CM_TRI_DEL:
        {
            Tri* t = glp_test_click_surface( view, &z, event->x, event->y );
            if( t == selected_tri )
                selected_tri = NULL;
            tris_del_v( current_mesh, t );
            t = NULL;
            rerender = TRUE;
        }
        break;


    default:
        break;
    }


    mouse_state[event->button].button_pressed = TRUE;
    mouse_state[event->button].has_moved = FALSE;
    mouse_state[event->button].startx = event->x;
    mouse_state[event->button].starty = event->y;
    mouse_state[event->button].endx = 0;
    mouse_state[event->button].endy = 0;
    mouse_state[event->button].view_index = view->type;


    if ( rerender ) {
        draw_all_frames();  // (in frame.c)
    }

    return TRUE;
}






gboolean view_mouse_motion( GtkWidget *widget, GdkEventMotion *event,
                           model_view *view ) {

    static Point * prev_coords = NULL;
//    static model_view* prev_view = NULL;
    int x, y;
    int btn = 0;
    Point * current_coords = NULL;
    GdkModifierType state;
    int redraw_current = FALSE;
    int redraw_all = FALSE;
    float deltaX, deltaY, deltaZ;

    g_return_val_if_fail( view != NULL, TRUE );

//    if( prev_view == NULL ) 
//        prev_view = view;
    
//printf( "prev_view is %i\n", prev_view->type );
//printf( "     view is %i\n", view->type );


    if ( event->is_hint )
        gdk_window_get_pointer ( event->window, &x, &y, &state );
    else {
        x = event->x;
        y = event->y;
        state = event->state;
    }
    


//    if ( !( state & GDK_BUTTON1_MASK || state & GDK_BUTTON2_MASK || state & GDK_BUTTON3_MASK ) )
//        return TRUE;



    if( x < 0 || x > view->glarea->allocation.width ||
        y < 0 || y > view->glarea->allocation.height )
        return TRUE;

    if( prev_coords == NULL ) {
        // first time we've been called.
        prev_coords = glp_test_click_ortho( view, x, y );
    }
    
    current_coords = glp_test_click_ortho( view, x, y );
    
    if( prev_coords == NULL || current_coords == NULL ) 
        return TRUE;
    
#if 0
    if( prev_view != view ) {
        prev_coords->v[0] = current_coords->v[0];
        prev_coords->v[1] = current_coords->v[1];
        prev_coords->v[2] = current_coords->v[2];
    }
#endif

    deltaX = current_coords->v[0] - prev_coords->v[0];
    deltaY = current_coords->v[1] - prev_coords->v[1];
    deltaZ = current_coords->v[2] - prev_coords->v[2];

    
/*    if ( grid_snap ) {
        a = grid_spacing * ( int ) rint( a / ( ( double ) grid_spacing ) );
        b = grid_spacing * ( int ) rint( b / ( ( double ) grid_spacing ) );
    }
*/

//printf( "in mouse_motion: changed_views is %i \n", changed_views );

    if( state & GDK_BUTTON1_MASK ) 
        btn = 1;
    else if( state & GDK_BUTTON2_MASK ) 
        btn = 2;
    else if( state & GDK_BUTTON3_MASK ) 
        btn = 3;


    if( !changed_views ) {
        switch( click_mode[ btn ] ) {
        
            case CM_POINT_MOVE:
                point_move_selected( deltaX, deltaY, deltaZ );
                redraw_all = TRUE;
                break;
                
            case CM_ZOOM:
                view_zoom( view, deltaX, deltaY, deltaZ );
                redraw_current = TRUE;
                break;
                
            case CM_PAN:
                view_pan( view, deltaX, deltaY, deltaZ );
                redraw_current = TRUE;
                break;
    
            default:
                break;
        }
        
        mouse_state[btn].has_moved = TRUE;
    
    } else if( btn == 0 ) {
        // if none of the buttons are being held down...
        changed_views = FALSE;
    }
    
    
    prev_coords->v[0] = current_coords->v[0];
    prev_coords->v[1] = current_coords->v[1];
    prev_coords->v[2] = current_coords->v[2];
    free( current_coords );
    

    if( redraw_all ) {
        draw_all_frames();
    } else if( redraw_current ) {
        draw_frame_n( view->type );
    }



    return TRUE;
}





gboolean view_mouse_up( GtkWidget *widget, GdkEventButton *event,
                           model_view *view ) {

    
    int btn;
    GLfloat z;
    int rerender = FALSE;
    Point* p = NULL;
    
    
    btn = event->button;
    
    
    mouse_state[btn].endx = event->x;
    mouse_state[btn].endy = event->y;
    
/*
printf( "mouse_state[%i] after btn release: \
\n\tpressed:\t%i \
\n\thas_movd:\t%i \
\n\tstartx:\t%i \
\tstarty:\t%i \
\n\tendx:\t%i \
\tendy:\t%i \n",
btn, mouse_state[btn].button_pressed, mouse_state[btn].has_moved, 
mouse_state[btn].startx, mouse_state[btn].starty,
mouse_state[btn].endx, mouse_state[btn].endy );
*/

    if( mouse_state[btn].button_pressed && current_mesh != NULL )
    switch( click_mode[ btn ] ) {
        
        case CM_POINT_SEL:
            
            if( mouse_state[btn].has_moved ) {
                point_unselect_all();
                view_lasso_select( view, mouse_state + btn );
                /*         pointer math -----------^   */
//FIXME - check that starting view is same as present view
            } else {
                p = glp_test_click_vertex( view, &z, event->x, event->y );

                if( p != NULL ) {
                    if( !p->selected )
                        // add the point to the selected list
                        point_select_v( p );
                    else
                        // remove the point from the selected list
                        point_unselect_v( p );
                } else {
                    point_unselect_all();
                }
            }
            rerender = TRUE;
        
            break;

        default:
            break;

    }



    mouse_state[btn].button_pressed = FALSE;
    mouse_state[btn].has_moved = FALSE;
    mouse_state[btn].startx = 0;
    mouse_state[btn].starty = 0;
    mouse_state[btn].endx = 0;
    mouse_state[btn].endy = 0;
    mouse_state[btn].view_index = -1;


    if ( rerender ) {
        draw_all_frames();  // (in frame.c)
    }

    return TRUE;
}







///////////////////////////////////////////////////////////////////////////
//
//
//  initialization stuff
//
//
///////////////////////////////////////////////////////////////////////////



model_view* view_new( int type ) {
    
    model_view* v = (model_view*)malloc( sizeof(model_view) );

    memset( v, 0, sizeof( model_view ) );
    
    v->type = type;
    view_initialize_config( v );
    
    return v;
}

void view_delete( GtkWidget *widget, GdkEvent *event, model_view *view ) {

    gtk_widget_destroy( view->glarea );
    gtk_widget_destroy( view->ghs );
    gtk_widget_destroy( view->gvs );
    gtk_widget_destroy( view->gltable );

    free( view );
}





GtkWidget *create_view( int type ) {

    GtkAdjustment * hadjustment, *vadjustment;
    GtkWidget *frame;

    gint attrlist[] = {
        GDK_GL_RGBA,
        GDK_GL_RED_SIZE, 1,
        GDK_GL_GREEN_SIZE, 1,
        GDK_GL_BLUE_SIZE, 1,
        GDK_GL_DOUBLEBUFFER,
        GDK_GL_NONE
    };

    views[type] = view_new( type );




//IMPORTANT?    gtk_signal_connect( GTK_OBJECT( views[type]->MainWindow ), "delete_event",
//IMPORTANT?                        GTK_SIGNAL_FUNC( view_delete ), views[type] );


    if ( gdk_gl_query() == FALSE ) {
        g_print( "OpenGL not available\n" );
        return NULL;
    }

    views[type]->glarea = gtk_gl_area_new( attrlist );

    gtk_widget_set_events( GTK_WIDGET( views[type]->glarea ),  
                           GDK_EXPOSURE_MASK|
                           GDK_BUTTON_PRESS_MASK|
                           GDK_BUTTON_RELEASE_MASK|
                           GDK_POINTER_MOTION_MASK|
                           GDK_POINTER_MOTION_HINT_MASK|
                           GDK_ENTER_NOTIFY_MASK|
                           GDK_LEAVE_NOTIFY_MASK );

    gtk_signal_connect( GTK_OBJECT( views[type]->glarea ), "realize",
                        GTK_SIGNAL_FUNC( view_glarea_realize ), views[type] );
    gtk_signal_connect( GTK_OBJECT( views[type]->glarea ), "configure_event",
                        GTK_SIGNAL_FUNC( view_glarea_configure ), views[type] );
    gtk_signal_connect( GTK_OBJECT( views[type]->glarea ), "expose_event",
                        GTK_SIGNAL_FUNC( view_glarea_expose ), views[type] );
//    gtk_signal_connect( GTK_OBJECT( views[type]->MainWindow ), "key_press_event",
//                        GTK_SIGNAL_FUNC( ed_key_press ), views[type] );
    gtk_signal_connect( GTK_OBJECT( views[type]->glarea ), "button_press_event",
                        GTK_SIGNAL_FUNC( view_mouse_down ), views[type] );
    gtk_signal_connect( GTK_OBJECT( views[type]->glarea ), "motion_notify_event",
                        GTK_SIGNAL_FUNC( view_mouse_motion ), views[type] );
    gtk_signal_connect( GTK_OBJECT( views[type]->glarea ), "button_release_event",
                        GTK_SIGNAL_FUNC( view_mouse_up ), views[type] );
    gtk_signal_connect( GTK_OBJECT( views[type]->glarea ), "leave_notify_event",
                        GTK_SIGNAL_FUNC( view_mouse_leave ), views[type] );

/*
  gint (* enter_notify_event)      (GtkWidget          *widget,
                                    GdkEventCrossing   *event);
  gint (* leave_notify_event)      (GtkWidget          *widget,
                                    GdkEventCrossing   *event);
*/

    frame = gtk_frame_new( NULL );
    gtk_frame_set_shadow_type( GTK_FRAME( frame ), GTK_SHADOW_OUT );
    gtk_container_add( GTK_CONTAINER( frame ), views[type]->glarea );
    gtk_widget_show( frame );

    views[type]->gltable = gtk_table_new( 2, 2, FALSE );

    gtk_table_attach( GTK_TABLE( views[type]->gltable ), frame, 0, 1, 0, 1,
                      GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0 );


    if( type == FRAME_3D ) {
        vadjustment = GTK_ADJUSTMENT( gtk_adjustment_new( 0, 0, 420, 5, 30, 60 ) );
        gtk_signal_connect ( GTK_OBJECT ( vadjustment ), "value_changed",
                             GTK_SIGNAL_FUNC ( vadj_changed ), NULL );
        views[type]->gvs = gtk_vscrollbar_new( vadjustment );
        gtk_widget_show( views[type]->gvs );
        gtk_table_attach( GTK_TABLE( views[type]->gltable ), views[type]->gvs, 1, 2, 0, 1,
                          GTK_SHRINK, GTK_EXPAND | GTK_FILL, 0, 0 );
    
        hadjustment = GTK_ADJUSTMENT( gtk_adjustment_new( 0, 0, 420, 5, 30, 60 ) );
        gtk_signal_connect ( GTK_OBJECT ( hadjustment ), "value_changed",
                             GTK_SIGNAL_FUNC ( hadj_changed ), NULL );
        views[type]->ghs = gtk_hscrollbar_new( hadjustment );
        gtk_widget_show( views[type]->ghs );
        gtk_table_attach( GTK_TABLE( views[type]->gltable ), views[type]->ghs, 0, 1, 1, 2,
                          GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 0 );
    }

    gtk_widget_show( views[type]->gltable );
    gtk_widget_show( views[type]->glarea );

    view_rebuild_glarrays( views[type] );

    gtk_widget_grab_focus( views[type]->glarea );


    return views[type]->gltable;
}





void initialize_mouse_state( void ) {
    
    int btn;
    
    for( btn = 0; btn < 3; btn++ ) {
        mouse_state[btn].button_pressed = FALSE;
        mouse_state[btn].has_moved = FALSE;
        mouse_state[btn].startx = 0;
        mouse_state[btn].starty = 0;
        mouse_state[btn].endx = 0;
        mouse_state[btn].endy = 0;
        mouse_state[btn].view_index = -1;
    }
}



/*
 *
 * view_initialize_config( ) - sets the view parameters to some 
 *     reasonable defaults
 *
 */

void view_initialize_config( model_view * view ) {

    
    if( view == NULL ) return;
//    if( view->config == NULL ) return;
    
    
    
    switch( view->type ) {
        
        case FRAME_XY:
            view->heading = 0.0;
            view->pitch = 0.0;
            view->dist = 10.0;

            view->grid_mins[ 0 ] = -20.0f;
            view->grid_mins[ 1 ] = -20.0f;
            view->grid_mins[ 2 ] = 0.0f;

            view->grid_maxes[ 0 ] = 20.0f;
            view->grid_maxes[ 1 ] = 20.0f;
            view->grid_maxes[ 2 ] = 0.0f;

            break;
        case FRAME_XZ:
            view->heading = 0.0;
            view->pitch = 90.0;
            view->dist = 10.0;

            view->grid_mins[ 0 ] = -20.0f;
            view->grid_mins[ 1 ] = 0.0f;
            view->grid_mins[ 2 ] = -20.0f;

            view->grid_maxes[ 0 ] = 20.0f;
            view->grid_maxes[ 1 ] = 0.0f;
            view->grid_maxes[ 2 ] = 20.0f;

            break;
        case FRAME_YZ:
            view->heading = 270.0;
            view->pitch = 90.0;
            view->dist = 10.0;

            view->grid_mins[ 0 ] = 0.0f;
            view->grid_mins[ 1 ] = -20.0f;
            view->grid_mins[ 2 ] = -20.0f;

            view->grid_maxes[ 0 ] = 0.0f;
            view->grid_maxes[ 1 ] = 20.0f;
            view->grid_maxes[ 2 ] = 20.0f;

            break;
        case FRAME_3D:
            view->heading = 0.0;
            view->pitch = 0.0;
            view->dist = 10.0;

            view->grid_mins[ 0 ] = -20.0f;
            view->grid_mins[ 1 ] = -20.0f;
            view->grid_mins[ 2 ] = 0.0f;

            view->grid_maxes[ 0 ] = 20.0f;
            view->grid_maxes[ 1 ] = 20.0f;
            view->grid_maxes[ 2 ] = 0.0f;

            break;
        default:
            break;
    }


    view->cursX = 0.0;
    view->cursY = 0.0;
    view->cursZ = 0.0;
    view->cursA = 0.0;


    view_config.fov = 90.0f;
    view_config.near = 1.0f;
    view_config.far = 200.0f;

    view_config.bkgnd_color[ 0 ] = 0.35f;
    view_config.bkgnd_color[ 1 ] = 0.35f;
    view_config.bkgnd_color[ 2 ] = 0.35f;

    view_config.grid_color[ 0 ] = 0.25f;
    view_config.grid_color[ 1 ] = 0.25f;
    view_config.grid_color[ 2 ] = 0.25f;

    view_config.grid_spacing[ 0 ] = 1.0f;
    view_config.grid_spacing[ 1 ] = 1.0f;
    view_config.grid_spacing[ 2 ] = 1.0f;

    view_config.vertices_visible = TRUE;
    view_config.grid_visible = TRUE;
    view_config.normals_visible = FALSE;
//    view_config.enlarge_vertices = FALSE;
    
    view_config.cursor_visible = FALSE;

    view_config.polygon_mode = MP_SHOW_WIREFRAME;
    view_config.triangle_order = MP_TRI_ORDER_CCW;

    view_config.cursor_rad = 0.2;   // normal cursor radius
//    view_config.cursor_rad = 0.6;   // enbiggened cursor radius

    view_config.vertex_rad = 0.1;   // normal vertex radius
//    view_config.vertex_rad = 0.5;   // enbiggened vertex radius

    view_config.girder_rad = 0.07;   // normal girder radius
//    view_config.girder_rad = 0.07;   // enbiggened girder radius

    initialize_mouse_state();
}




void view_reconfigure( model_view *view, gint flags ) {
    g_return_if_fail( view != NULL );
    if ( flags & RCV_PROPERTIES )
        glp_glconfigure( view );
    if ( flags & RCV_VIEWPORT ) {
        glp_glviewport( view );
        glp_ortho_viewport( view );
    }
}


void view_rebuild_glarrays(  ) {
    int i;

    if ( cursor_glarray != NULL )
        free( cursor_glarray );
    if ( vertex_glarray != NULL )
        free( vertex_glarray );
    if ( girder_glarray != NULL )
        free( girder_glarray );

    if ( view_config.triangle_order == MP_TRI_ORDER_CW ) {
        cursor_glarray = ( GLfloat * ) malloc( sizeof( view_cursor_cw ) );
        memcpy( cursor_glarray, view_cursor_cw, sizeof( view_cursor_cw ) );
        vertex_glarray = ( GLfloat * ) malloc( sizeof( view_vertex_cw ) );
        memcpy( vertex_glarray, view_vertex_cw, sizeof( view_vertex_cw ) );
        girder_glarray = ( GLfloat * ) malloc( sizeof( view_girder_cw ) );
        memcpy( girder_glarray, view_girder_cw, sizeof( view_girder_cw ) );
    } else {
        cursor_glarray = ( GLfloat * ) malloc( sizeof( view_cursor_ccw ) );
        memcpy( cursor_glarray, view_cursor_ccw, sizeof( view_cursor_ccw ) );
        vertex_glarray = ( GLfloat * ) malloc( sizeof( view_vertex_ccw ) );
        memcpy( vertex_glarray, view_vertex_ccw, sizeof( view_vertex_ccw ) );
        girder_glarray = ( GLfloat * ) malloc( sizeof( view_girder_ccw ) );
        memcpy( girder_glarray, view_girder_ccw, sizeof( view_girder_ccw ) );
    }


    /* size the arrays to fit the view_config.* parameters */
    for ( i = 0; i < ( sizeof( view_cursor_ccw ) / sizeof( GLfloat ) ); i += 10 ) {
        cursor_glarray[ i + 7 ] *= view_config.cursor_rad;
        cursor_glarray[ i + 8 ] *= view_config.cursor_rad;
        cursor_glarray[ i + 9 ] *= view_config.cursor_rad;
    }
    for ( i = 0; i < ( sizeof( view_vertex_ccw ) / sizeof( GLfloat ) ); i += 6 ) {
        vertex_glarray[ i + 3 ] *= view_config.vertex_rad;
        vertex_glarray[ i + 4 ] *= view_config.vertex_rad;
        vertex_glarray[ i + 5 ] *= view_config.vertex_rad;
    }
    for ( i = 0; i < ( sizeof( view_girder_ccw ) / sizeof( GLfloat ) ); i += 6 ) {
        girder_glarray[ i + 3 ] *= view_config.girder_rad;
        girder_glarray[ i + 4 ] *= view_config.girder_rad;
    }
}



///////////////////////////////////////////////////////////////////////
//
// other stuff
//
///////////////////////////////////////////////////////////////////////


void view_pan( model_view* view, float deltaX, float deltaY, float deltaZ ) {
    
    if( view != NULL ) {
        view->camX += deltaX;
        view->camY += deltaY;
        view->camZ += deltaZ;

        view->grid_mins[0] = -20.0f - view->camX;
        view->grid_mins[1] = -20.0f - view->camY;
//        view->grid_mins[2] = ;

        view->grid_maxes[0] = 20.0f - view->camX;
        view->grid_maxes[1] = 20.0f - view->camY;
//        view->grid_maxes[2] = ;
    }
}

void view_zoom( model_view* view, float deltaX, float deltaY, float deltaZ ) {
    
    if( view != NULL ) {
        view->dist -= (deltaX + deltaY + deltaZ);  // a hack...
        if( view->type != FRAME_3D )
            glp_ortho_viewport( view );
    }
}


int isinbounds( Point* p, Point* start, Point* end ) {
    /* checks if a point is within a specified region of space */
    
    
    float s[3];
    float e[3];
    int i;
    int result = TRUE;
    
    if( p == NULL || start == NULL || end == NULL )
        return FALSE;
    
    for( i = 0; i < 3; i++ ) {
        if( start->v[i] < end->v[i] ) {
            s[i] = start->v[i];
            e[i] = end->v[i];
        } else {
            s[i] = end->v[i];
            e[i] = start->v[i];
        }

        if( fabs( e[i] - s[i] ) > 0.01 )
            // if the starting and ending points on this axis are really really close together,
            // then it's likely that this axis is not one of the two axes represented by the current 
            // orthographic view (ie glp_test_click_ortho puts 0's in the point's unused axis,
            // so if, for example, s[1] and e[1] are both 0, we're probably dealing with 
            // the FRAME_XZ view)
        
            if( p->v[i] < s[i] || p->v[i] > e[i] )
                result = FALSE;

    }
    
    return result;
}



void view_lasso_select( model_view *view, struct _mouse_state* ms ) {
    /* called after buttondown, drag, buttonup */
    
    Point* start_point = NULL;
    Point* end_point = NULL;
    
    GList* l = NULL;


    start_point = glp_test_click_ortho( view, ms->startx, ms->starty );
    end_point = glp_test_click_ortho( view, ms->endx, ms->endy );

    // see note in glp_test_click_ortho
    start_point->v[0] -= view->camX;  end_point->v[0] -= view->camX;
    start_point->v[1] -= view->camY;  end_point->v[1] -= view->camY;
    start_point->v[2] -= view->camZ;  end_point->v[2] -= view->camZ;

    if( current_mesh->location != NULL ) {
        start_point->v[0] -= current_mesh->location->v[0];
        start_point->v[1] -= current_mesh->location->v[1];
        start_point->v[2] -= current_mesh->location->v[2];
        end_point->v[0] -= current_mesh->location->v[0];
        end_point->v[1] -= current_mesh->location->v[1];
        end_point->v[2] -= current_mesh->location->v[2];
    }

    
    for( l = current_mesh->points; l != NULL; l = l->next ) {
    
        if( isinbounds( (Point*)l->data, start_point, end_point ) ) {
            
            point_select_v( (Point*)l->data );
        }

    }
        
    
}




Point *glp_test_click_ortho( model_view *view, guint x, guint y ) {

// this is used *only* for zooming, panning, adding points, & lasso selection.
// presently, it only returns relative values (ie if you pan a view, 
// and then click in the middle of the view, the value returned by this func
// will be <0, 0, 0> not <camX, camY, camZ> )

// this func only exists because i can't figure out how to do 
// ray-plane intersections in opengl.  anybody care to help?

    Point* p = NULL;
    float win_delta_x, win_delta_y;
    float ortho_delta_x, ortho_delta_y, ortho_delta_z;
    
    g_return_val_if_fail ( view != NULL, NULL );
    g_return_val_if_fail ( view->glarea != NULL, NULL );
//    g_return_val_if_fail ( view->type != FRAME_3D, NULL );
    
    p = point_new();
    y = view->glarea->allocation.height - y;  // flip y axis
    
    win_delta_x = (float)x / (float)view->glarea->allocation.width;
    win_delta_y = (float)y / (float)view->glarea->allocation.height;
    
    ortho_delta_x = view->maxX - view->minX;
    ortho_delta_y = view->maxY - view->minY;
    ortho_delta_z = view->maxZ - view->minZ;
    
    switch( view->type ) {
        case FRAME_3D:         /* <--- used only for zoom & pan stuff... doesn't actually return sensible values */
        case FRAME_XY:
            p->v[0] = (ortho_delta_x * win_delta_x) + view->minX;
            p->v[1] = (ortho_delta_y * win_delta_y) + view->minY;
            break;
        case FRAME_XZ:
            p->v[0] = (ortho_delta_x * win_delta_x) + view->minX;
            p->v[2] = (ortho_delta_z * win_delta_y) + view->minZ;
            break;
        case FRAME_YZ:
            p->v[1] = (ortho_delta_y * win_delta_x) + view->minY;
            p->v[2] = (ortho_delta_z * win_delta_y) + view->minZ;
            break;
        default:
            break;
    }
    

//printf( "view->minX:  %f      view->minY:  %f      view->minZ:  %f\n", view->minX, view->minY, view->minZ );
//printf( "orthotest   x: %f    y: %f    z: %f\n", p->v[0], p->v[1], p->v[2]  );
    
    return p;
}



void view_set_rendermode( int mode ) {

    view_config.polygon_mode = mode;


}

void view_set_triangleorder( int order ) {
    
    int curr_view;
    
    for( curr_view = 0; curr_view < 4; curr_view++ ) {
    
        gtk_gl_area_make_current( GTK_GL_AREA( views[curr_view]->glarea ) );
        
        switch( order ) {
        
            case MP_TRI_ORDER_CCW:
                view_config.triangle_order = MP_TRI_ORDER_CCW;
                glEnable( GL_CULL_FACE );
                glFrontFace( GL_CCW );
                break;
    
            case MP_TRI_ORDER_CW:
                view_config.triangle_order = MP_TRI_ORDER_CW;
                glEnable( GL_CULL_FACE );
                glFrontFace( GL_CW );
                break;
    
            case MP_TRI_ORDER_ALL:
                view_config.triangle_order = MP_TRI_ORDER_ALL;
                glDisable( GL_CULL_FACE );
                break;
            
            default:
                break;
        }
    
    }

    view_rebuild_glarrays(  );


}

void view_set_vertexsize( float size ) {


/*
    switch( size ) {
        case MP_VERT_SIZE_OFF:
            view_config.vertices_visible = FALSE;
            break;

        case MP_VERT_SIZE_NORMAL:
            view_config.vertices_visible = TRUE;
            view_config.enlarge_vertices = FALSE;
            view_rebuild_glarrays(  );
            break;

        case MP_VERT_SIZE_BIG:
            view_config.vertices_visible = TRUE;
            view_config.enlarge_vertices = TRUE;
            view_rebuild_glarrays(  );
            break;
        
        default:
            break;
    }
*/


    view_config.vertex_rad = size;
    view_config.cursor_rad = size * 2.0;
    view_config.girder_rad = size * 0.7;

    view_rebuild_glarrays(  );

}


void view_set_fov( float fov ) {

    view_config.fov = fov;
    
    // FRAME_3D is the only view that uses the fov setting...
    view_reconfigure( views[FRAME_3D], RCV_VIEWPORT );


}








